/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <string.h>
#include "common.h"

#undef DEBUG

#if   defined(NN)
#define GEMM_LOCAL    GEMM_NN
#elif defined(NT)
#define GEMM_LOCAL    GEMM_NT
#elif defined(NR)
#define GEMM_LOCAL    GEMM_NR
#elif defined(NC)
#define GEMM_LOCAL    GEMM_NC
#elif defined(TN)
#define GEMM_LOCAL    GEMM_TN
#elif defined(TT)
#define GEMM_LOCAL    GEMM_TT
#elif defined(TR)
#define GEMM_LOCAL    GEMM_TR
#elif defined(TC)
#define GEMM_LOCAL    GEMM_TC
#elif defined(RN)
#define GEMM_LOCAL    GEMM_RN
#elif defined(RT)
#define GEMM_LOCAL    GEMM_RT
#elif defined(RR)
#define GEMM_LOCAL    GEMM_RR
#elif defined(RC)
#define GEMM_LOCAL    GEMM_RC
#elif defined(CN)
#define GEMM_LOCAL    GEMM_CN
#elif defined(CT)
#define GEMM_LOCAL    GEMM_CT
#elif defined(CR)
#define GEMM_LOCAL    GEMM_CR
#elif defined(CC)
#define GEMM_LOCAL    GEMM_CC
#endif

#ifndef SWITCH_RATIO
#define SWITCH_RATIO 2
#endif

typedef struct {
  long n, offset;

  volatile FLOAT *b;
  volatile FLOAT *buffer[2];

  volatile int working[MAX_CPU_NUMBER][16] __attribute__ ((aligned (128)));
} job_t;

static int inner_thread(long m, long n, long k,
			FLOAT alpha_r,
#ifdef COMPLEX
			FLOAT alpha_i,
#endif
			FLOAT *a, long lda,
			FLOAT *b, long ldb,
			FLOAT *c, long ldc, FLOAT *sa, FLOAT *sb,
			job_t *job, long mypos){

  long ls, min_l;
  long is, min_i;

  long i, buffer_side, current;

#ifdef DEBUG
  fprintf(stderr, "Thread Information[%ld] : m = %ld  n = %ld  k = %ld offset = %ld\n", 
	  mypos, m, n, k, offset);
#endif

  job[mypos].buffer[0] = sb;
  job[mypos].buffer[1] = sb + GEMM_Q * ((n + GEMM_UNROLL_N - 1) & ~(GEMM_UNROLL_N - 1)) * COMPSIZE;

  buffer_side = 0;

#ifdef DEBUG
  fprintf(stderr, "Thread Information[%ld] : Initilization is done\n", mypos);
#endif

  for(ls = 0; ls < k; ls += min_l){
    min_l = k - ls;
    if (min_l >= GEMM_Q * 2) {
      min_l = GEMM_Q;
    } else 
      if (min_l > GEMM_Q) {
	min_l = (min_l + 1) / 2;
      }

#ifdef DEBUG
    fprintf(stderr, "Thread Information[%ld] : Checking if job is finished...\n", mypos);
#endif

    for (i = 0; i < blas_cpu_number; i++) {
      while (job[mypos].working[i][buffer_side]) {};
    }
    
#ifdef DEBUG
  fprintf(stderr, "Thread Information[%ld] : Job has been finished.  LS = %ld Buffer = %p\n",
	  mypos, ls, job[mypos].buffer[buffer_side]);
#endif

#if defined(NN) || defined(TN) || defined(CN) || defined(RN) || \
    defined(NR) || defined(TR) || defined(CR) || defined(RR)
    GEMM_ONCOPY(min_l, n, (FLOAT *)b + (ls      ) * COMPSIZE, ldb, (FLOAT *)job[mypos].buffer[buffer_side]);
#else
    GEMM_OTCOPY(min_l, n, (FLOAT *)b + (ls * ldb) * COMPSIZE, ldb, (FLOAT *)job[mypos].buffer[buffer_side]);
#endif


#ifdef DEBUG
  fprintf(stderr, "Thread Information[%ld] : Copy has been done.\n", mypos);
#endif

  /* now other threads can access buffer */
  for (i = 0; i < blas_cpu_number; i++) job[mypos].working[i][buffer_side] = 1;

  for(is = 0; is < m; is += min_i){
    min_i = m - is;
    if (min_i >= GEMM_P * 2) {
      min_i = GEMM_P;
    } else 
      if (min_i > GEMM_P) {
	min_i = ((min_i + 1) / 2 + GEMM_UNROLL_M - 1) & ~(GEMM_UNROLL_M - 1);
      }
    
#if defined(NN) || defined(NT) || defined(NC) || defined(NR) || \
    defined(RN) || defined(RT) || defined(RC) || defined(RR)
      GEMM_ITCOPY(min_l, min_i, a + (is + ls * lda) * COMPSIZE, lda, sa);
#else
      GEMM_INCOPY(min_l, min_i, a + (ls + is * lda) * COMPSIZE, lda, sa);
#endif
    
    current = mypos;
    do {

      /* thread has to wait only at first time */
      if ((is == 0) && (current != mypos)) {
	while(job[current].working[mypos][buffer_side] == 0) {};
      }
      
#ifdef DEBUG
      fprintf(stderr, "Thread Information[%ld  --> %ld] : M = %ld, N = %ld K = %ld  offset = %ld.\n",
	      mypos, current, m, job[current].n, min_l, job[current].offset);
#endif
      
#if defined(NN) || defined(NT) || defined(TN) || defined(TT)
      GEMM_KERNEL_N
#endif
#if defined(CN) || defined(CT) || defined(RN) || defined(RT)
      GEMM_KERNEL_L
#endif
#if defined(NC) || defined(TC) || defined(NR) || defined(TR)
      GEMM_KERNEL_R
#endif
#if defined(CC) || defined(CR) || defined(RC) || defined(RR)
      GEMM_KERNEL_B
#endif
      (min_i, job[current].n, min_l,
       alpha_r,
#ifdef COMPLEX
       alpha_i,
#endif
       sa, (FLOAT *)job[current].buffer[buffer_side], 
       c + (is + job[current].offset * ldc) * COMPSIZE, ldc); 

#ifdef DEBUG
  fprintf(stderr, "Thread Information[%ld] : Finished.\n", mypos);
#endif

  if (is + GEMM_P >= m) {
    /* Thread doesn't need this buffer any more */
    job[current].working[mypos][buffer_side] = 0;
  }
  
  current ++;
  if (current >= blas_cpu_number) current = 0;
  
    } while (current != mypos);
  } /* end of is */
  
  
  buffer_side = 1 - buffer_side;
  
  }
  
  for (i = 0; i < blas_cpu_number; i++) {
      while (job[mypos].working[i][0] || job[mypos].working[i][1]) {};
  }

  return 0;
}

int CNAME(long m, long n, long k, 
	  FLOAT alpha_r,
#ifdef COMPLEX
	  FLOAT alpha_i,
#endif
	  FLOAT *a, long lda,
	  FLOAT *b, long ldb, 
	  FLOAT *c, long ldc, FLOAT *sa, FLOAT *sb){

  blas_queue_t queue[MAX_CPU_NUMBER];
  job_t        job  [MAX_CPU_NUMBER];

  long width, offset, i, j;
  int num_cpu, mode;

#ifndef COMPLEX
#ifdef XDOUBLE
  mode  =  BLAS_XDOUBLE | BLAS_REAL;
#elif defined(DOUBLE)
  mode  =  BLAS_DOUBLE  | BLAS_REAL;
#else
  mode  =  BLAS_SINGLE  | BLAS_REAL;
#endif  
#else
#ifdef XDOUBLE
  mode  =  BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
  mode  =  BLAS_DOUBLE  | BLAS_COMPLEX;
#else
  mode  =  BLAS_SINGLE  | BLAS_COMPLEX;
#endif  
#endif


#if defined(TN) || defined(TT) || defined(TR) || defined(TC) || \
    defined(CN) || defined(CT) || defined(CR) || defined(CC)
  mode |= (BLAS_TRANSA_T);
#endif
#if defined(NT) || defined(TT) || defined(RT) || defined(CT) || \
    defined(NC) || defined(TC) || defined(RC) || defined(CC)
  mode |= (BLAS_TRANSB_T);
#endif
  
  if (GEMM_P * blas_cpu_number > m * SWITCH_RATIO) {
    int divA, divB;

    divA = blas_cpu_number;
    divB = 1;

    while ((GEMM_P * divA > m * 2) && (divA > 1)) {
      do {
      divA --;
      divB = 1;
      while (divA * divB < blas_cpu_number) divB ++;
      } while ((divA * divB != blas_cpu_number) && (divA > 1));
    }

    gemm_thread_variable(divA, divB, mode, m, n, k, alpha_r,
#ifndef COMPLEX
			 ZERO,
#else
			 alpha_i,
#endif
			 a, lda, b, ldb, c, ldc, 
			 (void *)GEMM_LOCAL, sa);

    return 0;
  }

  for (i = 0; i < blas_cpu_number; i++){
    blas_queue_init(&queue[i]);

    queue[i].mode = mode;
    queue[i].routine = inner_thread;
    queue[i].k   = k;
    queue[i].lda = lda;
    queue[i].ldb = ldb;
    queue[i].ldc = ldc;
    queue[i].alpha[0] = alpha_r;
#ifndef COMPLEX
    queue[i].alpha[1] = ZERO;
#else
    queue[i].alpha[1] = alpha_i,
#endif
    queue[i].to    = i;
    queue[i].next  = &queue[i + 1];

  }
  queue[blas_cpu_number - 1].next = NULL;
  queue[blas_cpu_number - 1].last = 1;


  /* Divide for M */

  num_cpu = 0;
  i = m;

  while (i > 0){
    width  = blas_quickdivide(i + blas_cpu_number - 1 - num_cpu, blas_cpu_number - num_cpu);

    i -= width;
    if (i < 0) width = width + i;
    
    queue[num_cpu].m = width;
    queue[num_cpu].a = (void *)a;
    queue[num_cpu].c = (void *)c;

#if defined(NN) || defined(NT) || defined(NR) || defined(NC) || \
    defined(RN) || defined(RT) || defined(RR) || defined(RC)
    a += width * COMPSIZE;
#else
    a += width * lda * COMPSIZE;
#endif
    
    c += width * COMPSIZE;
    num_cpu ++;
  };

  /* Divide for N */

  num_cpu = 0;
  offset = 0;
  i = n;

  while (num_cpu < blas_cpu_number){
    if (i > 0) {
      width = blas_quickdivide(i + blas_cpu_number - 1 - num_cpu, blas_cpu_number - num_cpu);
      width = (width + GEMM_UNROLL_N - 1) & ~(GEMM_UNROLL_N - 1);
      i -= width;
      if (i < 0) width = width + i;
    } else {
      width = 0;
    }
    
    queue[num_cpu].n    = width;
    queue[num_cpu].b    = (void *)b;
    queue[num_cpu].from = (long)job;

    job[num_cpu].n      = width;
    job[num_cpu].b	= (void *)b;
    job[num_cpu].offset = offset;

    for (j = 0; j < blas_cpu_number; j++) {
      job[num_cpu].working[j][0] = 0;
      job[num_cpu].working[j][1] = 0;
    }

    offset += width;

#if defined(NN) || defined(TN) || defined(RN) || defined(CN) || \
    defined(NR) || defined(TR) || defined(RR) || defined(CR)
    b += width * ldb * COMPSIZE;
#else
    b += width * COMPSIZE;
#endif
    num_cpu ++;
  };

  exec_blas(queue, sa);

  return 0;
}
