/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "common.h"

int CNAME(int divideA, int divideB,
	  int mode, long m, long n, long k, 
	  double alpha_r, double alpha_i,
	  void *a, long lda,
	  void *b, long ldb, 
	  void *c, long ldc, int (*function)(), void *buffer){

  blas_queue_t queue[MAX_CPU_NUMBER];
  long width, i;
  int num_cpu, calc_type, cpus;

  long j, offset;
  long   sizeA[MAX_CPU_NUMBER];
  long offsetA[MAX_CPU_NUMBER];
  long   sizeB[MAX_CPU_NUMBER];
  long offsetB[MAX_CPU_NUMBER];

  calc_type = (mode & BLAS_PREC) + ((mode & BLAS_COMPLEX) != 0) + 2;
  
  for (i = 0; i < blas_cpu_number; i++) blas_queue_init(&queue[i]);

  if (divideA * divideB > blas_cpu_number) {
    fprintf(stderr, "BLAS thread warning : divide A = %d  divide B = %d is larger than actual number of threads\n", divideA, divideB);
    divideA = 1;
    divideB = blas_cpu_number;
  }

  /* divide rule for A */
  i = m;
  num_cpu = 0;
  offset = 0;
  while (i > 0){
    width  = blas_quickdivide(i + divideA - num_cpu - 1, divideA - num_cpu);

    if (width < GEMM_UNROLL_M * GEMM_UNROLL_M) width = GEMM_UNROLL_M * GEMM_UNROLL_M;

    i -= width;
    if (i < 0) width = width + i;
    sizeA  [num_cpu] = width;
    offsetA[num_cpu] = offset;
    offset += width;
    num_cpu ++;
  }
  divideA = num_cpu;

  cpus = 0;

  /* divide rule for B */
  i = n;
  num_cpu = 0;
  offset = 0;
  while (i > 0){
    width  = blas_quickdivide(i + divideB - num_cpu - 1, divideB - num_cpu);
    
    if (width < GEMM_UNROLL_N * GEMM_UNROLL_N) width = GEMM_UNROLL_N * GEMM_UNROLL_N;
    
    i -= width;
    if (i < 0) width = width + i;
    sizeB  [num_cpu] = width;
    offsetB[num_cpu] = offset;
    offset += width;
    num_cpu ++;
  }
  divideB = num_cpu;
  
  for (j = 0; j < divideB; j ++){
    for (i = 0; i < divideA; i ++){
      
      queue[cpus].mode = mode;
      queue[cpus].routine = function;
      queue[cpus].m = sizeA[i];
      queue[cpus].n = sizeB[j];
      queue[cpus].k = k;
      
      if (!(mode & BLAS_TRANSA_T)){
	queue[cpus].a = (void *)((long)a + (offsetA[i] << calc_type));
      } else {
	queue[cpus].a = (void *)((long)a + (offsetA[i] * lda << calc_type));
      }
      
      if (!(mode & BLAS_TRANSB_T)){
	queue[cpus].b = (void *)((long)b + (offsetB[j] * ldb << calc_type));
      } else {
	queue[cpus].b = (void *)((long)b + (offsetB[j] << calc_type));
      }

      queue[cpus].c = (void *)((long)c + ((offsetA[i] + ldc * offsetB[j]) << calc_type));
      
      queue[cpus].lda = lda;
      queue[cpus].ldb = ldb;
      queue[cpus].ldc = ldc;
      queue[cpus].alpha[0] = alpha_r;
      queue[cpus].alpha[1] = alpha_i;
      queue[cpus].next  = &queue[cpus + 1];
      
      cpus ++;
    }
  }

  if (cpus) {
    queue[cpus - 1].next = NULL;
    queue[cpus - 1].last = 1;
    
    exec_blas(queue, buffer);

  }
   
  return 0;
}
